package org.maltparser.core.config;

import java.io.File;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Formatter;


import org.apache.log4j.Logger;
import org.maltparser.MaltConsoleEngine;
import org.maltparser.core.dataformat.ColumnDescription;
import org.maltparser.core.dataformat.DataFormatSpecification;
import org.maltparser.core.dataformat.reader.DataReader;
import org.maltparser.core.dataformat.writer.DataWriter;
import org.maltparser.core.exception.MaltChainedException;
import org.maltparser.core.graph.DependencyGraph;
import org.maltparser.core.graph.SingleHeadedDependencyGraph;
import org.maltparser.core.options.OptionException;
import org.maltparser.core.sentence.Sentence;
import org.maltparser.core.symbol.SymbolTable;
import org.maltparser.core.symbol.SymbolTableHandler;
import org.maltparser.core.symbol.trie.TrieSymbolTableHandler;
import org.maltparser.parser.algorithm.ParsingAlgorithm;
import org.maltparser.parser.algorithm.helper.ParserAction;
import org.maltparser.parser.algorithm.nivre.malt04.NivreEagerMalt04;
import org.maltparser.parser.algorithm.nivre.malt04.NivreStandardMalt04;
import org.maltparser.parser.guide.Guidable;
import org.maltparser.parser.guide.Guide;

/**
 *  SingleMaltConfiguration handles deterministic parsing algorithms such as Nivre Arc-eager, 
 *  Nivre Arc-standard, Covington Non-projective and Covington projective. It is possible to run
 *  the Single Malt configuration in six different modes:
 *  
 *  <ul>
 *  <li>LEARN: Creates a configuration and induces a parsing model from input data.
 *  <li>PARSE: Parses sentences using a configuration.
 *  <li>UNPACK: Unpacks a configuration into a directory with the same name.
 *  <li>INFO: Prints information about a configuration.
 *  <li>PROJ: Creates a configuration and projectivizes input data without inducing a parsing model.
 *  <li>DEPROJ: Deprojectivizes input data using a configuration.
 *	</ul>
 *
 *	This class also implements the interface Guidable, which means that it should provide guidance to
 *	the parsing algorithm when requested and collect training instances during oracle parsing. In 
 *	the current implementation of single malt configuration it redirects these duties to a guide model.
 *
 * @author Johan Hall
 * @since 1.0
 * 
 * @see org.maltparser.core.config.Configuration
**/
public class SingleMaltConfiguration implements Configuration, Guidable {
	public static final int LEARN = 0;
	public static final int PARSE = 1;
	public static final int UNPACK = 2;
	public static final int INFO = 3;
	public static final int PROJ = 4;
	public static final int DEPROJ = 5;
	
	protected Logger configLogger;
	protected ConfigurationDir configDir;
	protected Integer containerIndex;
	protected ParsingAlgorithm parsingAlgorithm = null;
	protected Guide guide = null;
	protected Sentence sentence;
	protected DataFormatSpecification inputFormat;
	protected DataFormatSpecification outputFormat;
	protected DataReader reader;
	protected DataWriter writer;
	protected int mode = -1;
	protected SymbolTableHandler symbolTables;
	
	protected long startTime;
	protected long endTime;
	protected SimpleDateFormat dateFormat;
	
	public SingleMaltConfiguration() throws MaltChainedException {
		startTime = System.currentTimeMillis();
	}

	/******************************** Configuration interface ********************************/
	public void initialize(Integer containerIndex) throws MaltChainedException {
		this.containerIndex = containerIndex;
		
		initMode();
		initConfigurationDir();
		if (mode == SingleMaltConfiguration.LEARN) {
			initConfigLogger();	
			checkOptionDependency();
			initInput();
			initParsingAlgorithm(); 

			if (!getOptionValue("malt0.4", "posset").toString().equalsIgnoreCase("")) {
				if (configLogger.isInfoEnabled()) {
					configLogger.info("Loading part-of-speech tagset '"+getOptionValue("malt0.4", "posset").toString()+"'...\n");
				}
				getSymbolTables().loadTagset(getOptionValue("malt0.4", "posset").toString(), "POSTAG", getOptionValue("input", "charset").toString(), ColumnDescription.INPUT, getOptionValue("singlemalt", "null_value").toString());
			}
			if (!getOptionValue("malt0.4", "cposset").toString().equalsIgnoreCase("")) {
				if (configLogger.isInfoEnabled()) {
					configLogger.info("Loading coarse-grained part-of-speech tagset '"+getOptionValue("malt0.4", "cposset").toString()+"'...\n");
				}
				getSymbolTables().loadTagset(getOptionValue("malt0.4", "cposset").toString(), "CPOSTAG", getOptionValue("input", "charset").toString(), ColumnDescription.INPUT, getOptionValue("singlemalt", "null_value").toString());
			}
			if (!getOptionValue("malt0.4", "depset").toString().equalsIgnoreCase("")) {
				if (configLogger.isInfoEnabled()) {
					configLogger.info("Loading dependency type tagset '"+getOptionValue("malt0.4", "depset").toString()+"'...\n");
				}
				getSymbolTables().loadTagset(getOptionValue("malt0.4", "depset").toString(), "DEPREL", getOptionValue("input", "charset").toString(), ColumnDescription.OUTPUT, getOptionValue("singlemalt", "null_value").toString(), getOptionValue("graph", "root_label").toString());
			}
			
			if ((Boolean)getOptionValue("malt0.4", "behavior").equals(true)) {
				// TODO fix Covington
				SymbolTable classes = getSymbolTables().getSymbolTable("AllCombinedClassTable");
				SymbolTable deprel = getSymbolTables().getSymbolTable("DEPREL");
				Character separator = getOptionValue("guide", "classitem_separator").toString().charAt(0);
				if (getOptionValueString("singlemalt", "parsing_algorithm").toString().equals("nivrestandard")) {
					classes.addSymbol("RE");
				} 
				for (int i = 1; i < deprel.getValueCounter(); i++) {
					if (!getOptionValue("graph", "root_label").toString().equals(deprel.getSymbolCodeToString(i))) {
						classes.addSymbol("RA"+separator+deprel.getSymbolCodeToString(i));
					}
				}
				for (int i = 1; i < deprel.getValueCounter(); i++) {
					if (!getOptionValue("graph", "root_label").toString().equals(deprel.getSymbolCodeToString(i))) {
						classes.addSymbol("LA"+separator+deprel.getSymbolCodeToString(i));
					}
				}
			}
		} else if (mode == SingleMaltConfiguration.PARSE) {
			initConfigLogger();	
			MaltConsoleEngine.getOptionManager().loadOptions(configDir.getInputStreamReader("savedoptions.sop"));
			initInput();
			if (configLogger.isDebugEnabled()) {
				configLogger.debug("Loading the symbol tables from 'symboltables.sym' ...\n");
			}
			getSymbolTables().load(configDir.getInputStreamReader("symboltables.sym",getOptionValue("input", "charset").toString()));
			initParsingAlgorithm();	
			initOutput();
			if (!getOptionValue("malt0.4", "depset").toString().equalsIgnoreCase("")) {
				if (configLogger.isInfoEnabled()) {
					configLogger.info("Loading dependency type tagset '"+getOptionValue("malt0.4", "depset").toString()+"'...\n");
				}
				getSymbolTables().loadTagset(getOptionValue("malt0.4", "depset").toString(),"DEPREL_DEPROJ", getOptionValue("input", "charset").toString(), ColumnDescription.OUTPUT, getOptionValue("singlemalt", "null_value").toString());
			}
		} else if (mode == SingleMaltConfiguration.PROJ) {
			initConfigLogger();
			initInput();
			initOutput();
		} else if (mode == SingleMaltConfiguration.DEPROJ) {
			initConfigLogger();
			MaltConsoleEngine.getOptionManager().loadOptions(configDir.getInputStreamReader("savedoptions.sop"));
			initInput();
			if (configLogger.isDebugEnabled()) {
				configLogger.debug("Loading the symbol tables from 'symboltables.sym' ...\n");
			}
			getSymbolTables().load(configDir.getInputStreamReader("symboltables.sym",getOptionValue("input", "charset").toString()));
			if (!getOptionValue("malt0.4", "depset").toString().equalsIgnoreCase("")) {
				if (configLogger.isInfoEnabled()) {
					configLogger.info("Loading dependency type tagset '"+getOptionValue("malt0.4", "depset").toString()+"'...\n");
				}
				getSymbolTables().loadTagset(getOptionValue("malt0.4", "depset").toString(),"DEPREL_DEPROJ", getOptionValue("input", "charset").toString(), ColumnDescription.OUTPUT, getOptionValue("singlemalt", "null_value").toString());
			}
			initOutput();
		}
	}
	
	/* (non-Javadoc)
	 * @see org.maltparser.core.config.Configuration#process()
	 */
	public void process() throws MaltChainedException {
		if (mode == SingleMaltConfiguration.LEARN) {
			learn();
		} else if (mode == SingleMaltConfiguration.PARSE) {
			parse();
		} else if (mode == SingleMaltConfiguration.INFO) {
			configDir.echoInfoFile();
		} else if (mode == SingleMaltConfiguration.PROJ) {
			projectivize();
		} else if (mode == SingleMaltConfiguration.DEPROJ) {
			deprojectivize();
		}
	}
	

	public void finalize() throws MaltChainedException {
		if (mode == SingleMaltConfiguration.LEARN) {
			configDir.createConfigFile();
			endTime = System.currentTimeMillis();
			long elapsed = endTime - startTime;
			if (configLogger.isInfoEnabled()) {
				configLogger.info("Finished: " + new Date(System.currentTimeMillis())+"\n");

				long nHours = elapsed / 3600000; 
				long nMinutes = elapsed % 3600000 / 60000; 
				long nSeconds = elapsed % 60000 / 1000; 
				configLogger.info("Learning time: " +new Formatter().format("%02d:%02d:%02d", nHours, nMinutes, nSeconds)+" ("+elapsed+" ms)\n");
			}
			configDir.finalize();
			configDir.deleteConfigDirectory();
		} else if (mode == SingleMaltConfiguration.PARSE) {
			endTime = System.currentTimeMillis();
			long elapsed = endTime - startTime;
			if (configLogger.isInfoEnabled()) {
				configLogger.info("Finished: " + new Date(System.currentTimeMillis())+"\n");
				long nHours = elapsed / 3600000; 
				long nMinutes = elapsed % 3600000 / 60000; 
				long nSeconds = elapsed % 60000 / 1000; 

				configLogger.info("Parsing time: " +new Formatter().format("%02d:%02d:%02d", nHours, nMinutes, nSeconds)+" ("+elapsed+" ms)\n");
			}
			configDir.deleteConfigDirectory();
		} else if (mode == SingleMaltConfiguration.INFO) {
			configDir.deleteConfigDirectory();
		} else if (mode == SingleMaltConfiguration.PROJ) {
 			configDir.createConfigFile();
			endTime = System.currentTimeMillis();
			long elapsed = endTime - startTime;
			if (configLogger.isInfoEnabled()) {
				configLogger.info("Finished: " + new Date(System.currentTimeMillis())+"\n");

				long nHours = elapsed / 3600000; 
				long nMinutes = elapsed % 3600000 / 60000; 
				long nSeconds = elapsed % 60000 / 1000; 
				configLogger.info("Projectivization time: "+new Formatter().format("%02d:%02d:%02d", nHours, nMinutes, nSeconds)+" ("+elapsed+" ms)\n");
			}
			configDir.finalize();
			configDir.deleteConfigDirectory();
		} else if (mode == SingleMaltConfiguration.DEPROJ) {
			endTime = System.currentTimeMillis();
			long elapsed = endTime - startTime;
			if (configLogger.isInfoEnabled()) {
				configLogger.info("Finished: " + new Date(System.currentTimeMillis())+"\n");
				long nHours = elapsed / 3600000; 
				long nMinutes = elapsed % 3600000 / 60000; 
				long nSeconds = elapsed % 60000 / 1000; 

				configLogger.info("Deprojectivization time: " +new Formatter().format("%02d:%02d:%02d", nHours, nMinutes, nSeconds)+" ("+elapsed+" ms)\n");
			}
			configDir.deleteConfigDirectory();
		}
		if (configLogger != null) {
			configLogger.removeAllAppenders();
		}
	}

	/* (non-Javadoc)
	 * @see org.maltparser.core.config.Configuration#getMode()
	 */
	public int getMode() {
		return mode;
	}
	
	/* (non-Javadoc)
	 * @see org.maltparser.core.config.Configuration#getOutputStreamWriter(java.lang.String)
	 */
	public OutputStreamWriter getOutputStreamWriter(String suffix) throws MaltChainedException {
		return configDir.getOutputStreamWriter(suffix, getOptionValue("output", "charset").toString());
	}
	
	/* (non-Javadoc)
	 * @see org.maltparser.core.config.Configuration#getInputStreamReader(java.lang.String)
	 */
	public InputStreamReader getInputStreamReader(String suffix) throws MaltChainedException {
		return configDir.getInputStreamReader(suffix, getOptionValue("input", "charset").toString());
	}
	
	/* (non-Javadoc)
	 * @see org.maltparser.core.config.Configuration#getFile(java.lang.String)
	 */
	public File getFile(String suffix) throws MaltChainedException {
		return configDir.getFile(suffix);
	}
	
	public BufferedWriter getInfoFileWriter() throws MaltChainedException {
		return configDir.getInfoFileWriter();
	}
	
	/* (non-Javadoc)
	 * @see org.maltparser.core.config.Configuration#getConfigLogger()
	 */
	public Logger getConfigLogger() {
		return configLogger;
	}
	
	/* (non-Javadoc)
	 * @see org.maltparser.core.config.Configuration#getName()
	 */
	public String getName() {
		return configDir.getName();
	}
	
	/* (non-Javadoc)
	 * @see org.maltparser.core.config.Configuration#getOptionValue(java.lang.String, java.lang.String)
	 */
	public Object getOptionValue(String optiongroup, String optionname) throws MaltChainedException {
		return MaltConsoleEngine.getOptionManager().getOptionValue(containerIndex, optiongroup, optionname);
	}
	
	protected String getOptionValueString(String optiongroup, String optionname) throws MaltChainedException {
		return MaltConsoleEngine.getOptionManager().getOptionValueString(containerIndex, optiongroup, optionname);
	}
	/******************************** Guidable interface ********************************/
	
	/**
	 * This method is used during learning. Currently, 
	 * the MaltParserConfiguration redirect the instance to the guide. 
	 * Maybe in the future this method will
	 * be re-implemented to add some interesting things or maybe not.
	 * 
	 * @param parserAction
	 * @throws MaltChainedException
	 */
	public void setInstance(ParserAction parserAction) throws MaltChainedException {
		if (mode != SingleMaltConfiguration.LEARN) {
			throw new ConfigurationException("It is only possible to set an instance during learning. ");
		}
		try {
			guide.addInstance(parserAction);
		} catch (NullPointerException e) {
			throw new ConfigurationException("The guide cannot be found. ", e);
		}
	}
	
	/**
	 * This method is used during parsing. Currently, 
	 * the MaltParserConfiguration redirect the request to the guide. 
	 * Maybe in the future this method will
	 * be re-implemented to add some interesting things or maybe not.
	 * 
	 * @throws MaltChainedException
	 */
	public boolean predictFromKBestList() throws MaltChainedException {
		try {
			return guide.predictFromKBestList();
		} catch (NullPointerException e) {
			throw new ConfigurationException("The guide cannot be found. ", e);
		}
	}
	
	/**
	 * This method is used during parsing. Currently, 
	 * the MaltParserConfiguration redirect the request to the guide. 
	 * Maybe in the future this method will
	 * be re-implemented to add some interesting things or maybe not.
	 * 
	 * @throws MaltChainedException
	 */
	public boolean predict() throws MaltChainedException {
		try {
			return guide.predict();
		} catch (NullPointerException e) {
			throw new ConfigurationException("The guide cannot be found. ", e);
		}
	}
	
	/******************************** MaltParserConfiguration specific  ********************************/
	
	/**
	 * Returns the list of symbol tables
	 * 
	 * @return the list of symbol tables
	 */
	public SymbolTableHandler getSymbolTables() {
		return symbolTables;
	}
	
	/**
	 * Returns the sentence
	 * 
	 * @return the sentence
	 */
	public Sentence getSentence() {
		return sentence;
	}
	
	/**
	 * Returns the parsing algorithm in use
	 * 
	 * @return the parsing algorithm in use
	 */
	public ParsingAlgorithm getParsingAlgorithm() {
		return parsingAlgorithm;
	}
	
	/******************************** MaltParserConfiguration helpers  ********************************/
	
	/**
	 * Learns how to parse sentences in a language. The data reader reads a sentence into 
	 * the sentence data structure. A dependency graph is build according to the gold standard. The
	 * gold standard graph is send to the oracle parser for a specific parsing algorithm.
	 *
	 * @throws MaltChainedException
	 */
	protected void learn() throws MaltChainedException {
		SingleHeadedDependencyGraph goldDependencyGraph = new SingleHeadedDependencyGraph(sentence, getOptionValue("graph", "root_label").toString());
		configLogger.info("Reading sentences from '"+getOptionValue("input", "infile").toString()+"':");
		int nSentences;
		while (reader.readSentence()) {
			if (sentence.hasTokens()) {	
				goldDependencyGraph.createFromGoldStandard();
				if (!getOptionValue("pproj", "marking_strategy").toString().equalsIgnoreCase("none")) { //|| !getOptionValue("pproj", "covered_root").toString().equalsIgnoreCase("none")) {
					goldDependencyGraph.projectivize(getOptionValue("pproj", "marking_strategy").toString(), getOptionValue("pproj", "covered_root").toString(), getOptionValue("pproj", "lifting_order").toString());
				}
				guide.finalizeSentence(sentence, parsingAlgorithm.oracleParse(goldDependencyGraph));
				nSentences = reader.getSentenceCount();
				if (nSentences == 1 || nSentences == 50 || nSentences == 100 || nSentences % 200 == 0) {
					configLogger.info(reader.getSentenceCount() + " ");
				}
			}
		}
		
		if (configLogger.isInfoEnabled()) {
			configLogger.info("\nNumber of sentences: "+reader.getSentenceCount()+"\n");
		}
		reader.close();

		guide.updateCardinality();
		if (configLogger.isInfoEnabled()) {
			configLogger.info("Creating all models\n");
		}

		guide.noMoreInstances();
		
		if (configLogger.isInfoEnabled()) {
			configLogger.info("Saving the symbol table...\n");
		}
		
		getSymbolTables().save(configDir.getOutputStreamWriter("symboltables.sym", getOptionValue("output", "charset").toString()));
		if (configLogger.isInfoEnabled()) {
			configLogger.info("Saving the configuration specific options...\n");
		}
		MaltConsoleEngine.getOptionManager().saveOptions(containerIndex, configDir.getOutputStreamWriter("savedoptions.sop"));
		guide.finalize();
	}
	
	/**
	 * Parses sentences from the data source. The data reader reads a sentence into 
	 * the sentence data structure and the parsing algorithm parses the sentence. Finally,
	 * the data writer writes to the data destination (often a file).
	 * 
	 * @throws MaltChainedException
	 */
	protected void parse() throws MaltChainedException {
		configLogger.info("Parsing sentences from '"+getOptionValue("input", "infile").toString()+"': ");
		int nSentences = 0;
		while (reader.readSentence()) {
			if (sentence.hasTokens()) {
				DependencyGraph parseDependencyGraph = parsingAlgorithm.parse(); 
				if (!getOptionValue("pproj", "marking_strategy").toString().equalsIgnoreCase("none")) { 
					parseDependencyGraph.deprojectivize(getOptionValue("pproj", "marking_strategy").toString());
				}
				writer.writeSentence(parseDependencyGraph);
				nSentences = reader.getSentenceCount();
				if (nSentences == 1 || nSentences == 10 || nSentences == 25 || nSentences == 50 || nSentences % 100 == 0) {
					configLogger.info(reader.getSentenceCount() + " ");
				}
			}
		}
		configLogger.info("\nNumber of parsed sentences: "+reader.getSentenceCount()+"\n");
		reader.close();
		writer.close();
	}
	
	
	protected void projectivize() throws MaltChainedException {
		SingleHeadedDependencyGraph goldDependencyGraph = new SingleHeadedDependencyGraph(sentence, getOptionValue("graph", "root_label").toString());
		configLogger.info("Projectivize sentences from '"+getOptionValue("input", "infile").toString()+"': ");
		int nSentences = 0;
		while (reader.readSentence()) {
			if (sentence.hasTokens()) {
				goldDependencyGraph.createFromGoldStandard();
				goldDependencyGraph.projectivize(getOptionValue("pproj", "marking_strategy").toString(), getOptionValue("pproj", "covered_root").toString(), getOptionValue("pproj", "lifting_order").toString());
				writer.writeSentence(goldDependencyGraph);
				nSentences = reader.getSentenceCount();
				if (nSentences == 1 || nSentences == 10 || nSentences == 25 || nSentences == 50 || nSentences % 100 == 0) {
					configLogger.info(reader.getSentenceCount() + " ");
				}
			}
		}
		configLogger.info("\nNumber of projectivized sentences: "+reader.getSentenceCount()+"\n");
		reader.close();
		writer.close();	
		getSymbolTables().save(configDir.getOutputStreamWriter("symboltables.sym", getOptionValue("output", "charset").toString()));
		if (configLogger.isInfoEnabled()) {
			configLogger.info("Saving the configuration specific options...\n");
		}
		MaltConsoleEngine.getOptionManager().saveOptions(containerIndex, configDir.getOutputStreamWriter("savedoptions.sop"));
	}
	
	protected void deprojectivize() throws MaltChainedException {
		SingleHeadedDependencyGraph goldDependencyGraph = new SingleHeadedDependencyGraph(sentence, getOptionValue("graph", "root_label").toString());
		configLogger.info("Deprojectivize sentences from '"+getOptionValue("input", "infile").toString()+"': ");
		int nSentences = 0;
		while (reader.readSentence()) {
			if (sentence.hasTokens()) {
				goldDependencyGraph.createFromGoldStandard();
				goldDependencyGraph.deprojectivize(getOptionValue("pproj", "marking_strategy").toString());
				writer.writeSentence(goldDependencyGraph);
				nSentences = reader.getSentenceCount();
				if (nSentences == 1 || nSentences == 10 || nSentences == 25 || nSentences == 50 || nSentences % 100 == 0) {
					configLogger.info(reader.getSentenceCount() + " ");
				}
			}
		}
		configLogger.info("\nNumber of projectivized sentences: "+reader.getSentenceCount()+"\n");
		reader.close();
		writer.close();	
	}
	
	/**
	 * Initialize the configuration mode
	 * 
	 * @throws MaltChainedException
	 */
	protected void initMode() throws MaltChainedException {
		if (getOptionValue("singlemalt", "mode") != null) {
			if (getOptionValue("singlemalt", "mode").toString().equalsIgnoreCase("parse")) {
				mode = SingleMaltConfiguration.PARSE;
			} else if (getOptionValue("singlemalt", "mode").toString().equalsIgnoreCase("learn")) {
				mode = SingleMaltConfiguration.LEARN;
			} else if (getOptionValue("singlemalt", "mode").toString().equalsIgnoreCase("unpack")) {
				mode = SingleMaltConfiguration.UNPACK;
			} else if (getOptionValue("singlemalt", "mode").toString().equalsIgnoreCase("info")) {
				mode = SingleMaltConfiguration.INFO;
			} else if (getOptionValue("singlemalt", "mode").toString().equalsIgnoreCase("proj")) {
				mode = SingleMaltConfiguration.PROJ;
			} else if (getOptionValue("singlemalt", "mode").toString().equalsIgnoreCase("deproj")) {
				mode = SingleMaltConfiguration.DEPROJ;
			}
		}
		if (mode == -1) {
			throw new ConfigurationException("Unknown configuration mode. ");
		}
	}
	
	/**
	 * Creates a configuration directory and assign the configuration a name. 
	 * 
	 */
	protected void initConfigurationDir() throws MaltChainedException {
		
		if (mode == SingleMaltConfiguration.LEARN) {
			configDir = new ConfigurationDir(this, getOptionValue("config", "name").toString(), "singlemalt", containerIndex);
			configDir.createConfigDirectory();
			String optionFileName = null;
			try {
				if (getOptionValue("system", "option_file") != null) {
					optionFileName = getOptionValue("system", "option_file").toString();
					if (optionFileName != null && optionFileName.length() > 0) {
						File optionFile = new File(getOptionValue("system", "option_file").toString());
						configDir.copyToConfig(optionFile);
					}
				}
			} catch (ConfigurationException e) {
				throw new ConfigurationException("The option file '"+optionFileName+"' could not be copied. ",e);
			}
		} else if (mode == SingleMaltConfiguration.PARSE || mode == SingleMaltConfiguration.UNPACK) {
			if (getOptionValue("config", "url") != null && getOptionValue("config", "url").toString().length() > 0) {
				try {
					configDir = new ConfigurationDir(this, new URL(getOptionValue("config", "url").toString()));
				} catch (MalformedURLException e) {
					throw new ConfigurationException("The URL '"+getOptionValue("config", "url").toString()+"' is malformed. ", e);
				}
			} else {
				configDir = new ConfigurationDir(this, getOptionValue("config", "name").toString(), "singlemalt", containerIndex);
			}
			configDir.unpackConfigFile();
		} else if (mode == SingleMaltConfiguration.INFO) {
			if (getOptionValue("config", "url") != null && getOptionValue("config", "url").toString().length() > 0) {
				try {
					configDir = new ConfigurationDir(this, new URL(getOptionValue("config", "url").toString()));
				} catch (MalformedURLException e) {
					throw new ConfigurationException("The URL '"+getOptionValue("config", "url").toString()+"' is malformed. ", e);
				}
			} else {
				configDir = new ConfigurationDir(this, getOptionValue("config", "name").toString(), "singlemalt", containerIndex);
			}		
		} else if (mode == SingleMaltConfiguration.PROJ) {
			configDir = new ConfigurationDir(this, getOptionValue("config", "name").toString(), "singlemalt", containerIndex);
			configDir.createConfigDirectory();
			String optionFileName = null;
			try {
				if (getOptionValue("system", "option_file") != null) {
					optionFileName = getOptionValue("system", "option_file").toString();
					if (optionFileName != null && optionFileName.length() > 0) {
						File optionFile = new File(getOptionValue("system", "option_file").toString());
						configDir.copyToConfig(optionFile);
					}
				}
			} catch (ConfigurationException e) {
				throw new ConfigurationException("The option file '"+optionFileName+"' could not be copied. ",e);
			}
		} else if (mode == SingleMaltConfiguration.DEPROJ) {
			if (getOptionValue("config", "url") != null && getOptionValue("config", "url").toString().length() > 0) {
				try {
					configDir = new ConfigurationDir(this, new URL(getOptionValue("config", "url").toString()));
				} catch (MalformedURLException e) {
					throw new ConfigurationException("The URL '"+getOptionValue("config", "url").toString()+"' is malformed. ", e);
				}
			} else {
				configDir = new ConfigurationDir(this, getOptionValue("config", "name").toString(), "singlemalt", containerIndex);
			}
			configDir.unpackConfigFile();
		}
	}
	
	protected void initConfigLogger() throws MaltChainedException {
		configLogger = configDir.initConfigLogger();
		
		if (configLogger.isInfoEnabled()) {
			configLogger.info(MaltConsoleEngine.header() +"\n");
			configLogger.info("Started: " + new Date(startTime) +"\n");
		}
	}
	
	/**
	 * Initialize the input. Creates a symbol table handler and a sentence. The data writer reads into 
	 * the sentence.
	 * 
	 * @throws MaltChainedException
	 */
	protected void initInput() throws MaltChainedException {
			inputFormat = new DataFormatSpecification(getConfigLogger());
			inputFormat.parseDataFormatXMLfile(getOptionValue("input", "format").toString());
			symbolTables = new TrieSymbolTableHandler(); 
			Integer sentenceCapacity = null;
			try {
				sentenceCapacity = Integer.parseInt(getOptionValue("graph", "max_sentence_length").toString());
			} catch (OptionException e) {
				if (configLogger.isDebugEnabled()) {
					configLogger.debug("The option 'maxsentencelength' isn't found (default value = "+Sentence.DEFAULTMAXSENLEN+" is used.\n");
				}
				sentenceCapacity = null;
			} catch (NumberFormatException e) {
				if (configLogger.isDebugEnabled()) {
					configLogger.debug("The option 'maxsentencelength' isn't an integer value (default value = "+Sentence.DEFAULTMAXSENLEN+" is used.\n");
				}
				sentenceCapacity = null;
			}			
			if (sentenceCapacity != null) {
				sentence = new Sentence(getConfigLogger(), inputFormat.createDataFormatInstance(symbolTables, getOptionValue("singlemalt", "null_value").toString(), getOptionValue("graph", "root_label").toString()), sentenceCapacity);
			} else {
				sentence = new Sentence(getConfigLogger(), inputFormat.createDataFormatInstance(symbolTables, getOptionValue("singlemalt", "null_value").toString(), getOptionValue("graph", "root_label").toString()));
			}
			
			
			Class<?> clazz = (Class<?>)getOptionValue("input", "reader");
			Class<?>[] argTypes = { org.maltparser.core.sentence.Sentence.class };
			Object[] arguments = new Object[1];
			arguments[0] = sentence;
			try {	
				Constructor<?> constructor = clazz.getConstructor(argTypes);
				reader = (DataReader)constructor.newInstance(arguments);
				reader.initialize(getOptionValue("input", "infile").toString(), getOptionValue("input", "charset").toString());
			} catch (NoSuchMethodException e) {
				throw new ConfigurationException("The data reader '"+clazz.getName()+"' cannot be initialized. ", e);
			} catch (InstantiationException e) {
				throw new ConfigurationException("The data reader '"+clazz.getName()+"' cannot be initialized. ", e);
			} catch (IllegalAccessException e) {
				throw new ConfigurationException("The data reader '"+clazz.getName()+"' cannot be initialized. ", e);
			} catch (InvocationTargetException e) {
				throw new ConfigurationException("The data reader '"+clazz.getName()+"' cannot be initialized. ", e);
			} 	
	}

	/**
	 * Initialize the output with a data writer
	 * 
	 * @throws MaltChainedException
	 */
	protected void initOutput() throws MaltChainedException {
		outputFormat = new DataFormatSpecification(getConfigLogger());
		outputFormat.parseDataFormatXMLfile(getOptionValue("input", "format").toString());
		if (!getOptionValue("input", "format").toString().equals(getOptionValue("output", "format").toString())) {
			configLogger.warn("Sorry! (Not implemented yet). It is not possible to have different values " +
					"of --input-format and --output-format. This will hopefully be possible in the future together with " +
					"the --inout-map option. For now, the MaltParser will use the --input-format as output format. ");
		}
		if (mode == SingleMaltConfiguration.PARSE || mode == SingleMaltConfiguration.PROJ || mode == SingleMaltConfiguration.DEPROJ) {
			Class<?> clazz = (Class<?>)getOptionValue("output", "writer");

			try {	
				writer = (DataWriter)clazz.newInstance();
				writer.initialize(getOptionValue("output", "outfile").toString(), getOptionValue("output", "charset").toString());
			} catch (InstantiationException e) {
				throw new ConfigurationException("The data writer '"+clazz.getName()+"' cannot be initialized. ", e);
			} catch (IllegalAccessException e) {
				throw new ConfigurationException("The data writer '"+clazz.getName()+"' cannot be initialized. ", e);
			}
		}
	}
	
	/**
	 * Initialize the parsing algorithm
	 * 
	 * @throws MaltChainedException
	 */
	protected void initParsingAlgorithm() throws MaltChainedException {
		if (((Boolean)getOptionValue("malt0.4", "behavior")).booleanValue() == true && getOptionValueString("singlemalt", "parsing_algorithm").equals("nivreeager")) {
			this.parsingAlgorithm = new NivreEagerMalt04(this, sentence);
		} else if (((Boolean)getOptionValue("malt0.4", "behavior")).booleanValue() == true && getOptionValueString("singlemalt", "parsing_algorithm").equals("nivrestandard")) {
			this.parsingAlgorithm = new NivreStandardMalt04(this, sentence);
		} else {
			Class<?> clazz = (Class<?>)getOptionValue("singlemalt", "parsing_algorithm");
	
			Class<?>[] argTypes = { org.maltparser.core.config.SingleMaltConfiguration.class, org.maltparser.core.sentence.Sentence.class };
			Object[] arguments = new Object[2];
			arguments[0] = this;
			arguments[1] = sentence;
			if (configLogger.isInfoEnabled()) {
				configLogger.info("Initialize the parsing algorithm...\n");
			}
			try {	
				Constructor<?> constructor = clazz.getConstructor(argTypes);
				this.parsingAlgorithm = (ParsingAlgorithm)constructor.newInstance(arguments);
			} catch (NoSuchMethodException e) {
				throw new ConfigurationException("The parsing algorithm '"+clazz.getName()+"' cannot be initialized. ", e);
			} catch (InstantiationException e) {
				throw new ConfigurationException("The parsing algorithm '"+clazz.getName()+"' cannot be initialized. ", e);
			} catch (IllegalAccessException e) {
				throw new ConfigurationException("The parsing algorithm '"+clazz.getName()+"' cannot be initialized. ", e);
			} catch (InvocationTargetException e) {
				throw new ConfigurationException("The parsing algorithm '"+clazz.getName()+"' cannot be initialized. ", e);
			}
		}
		guide = new Guide(this, parsingAlgorithm);
	}
	
	
	
	public void checkOptionDependency() throws MaltChainedException {
		try {
			configDir.getInfoFileWriter().write("\nDEPENDENCIES\n");
			if ((Boolean)getOptionValue("malt0.4", "behavior") == true) {
				if (!getOptionValueString("singlemalt", "null_value").equals("rootlabel")) {
					MaltConsoleEngine.getOptionManager().overloadOptionValue(containerIndex, "singlemalt", "null_value", "rootlabel");
					configDir.getInfoFileWriter().write("--singlemalt-null_value (-nv)     rootlabel\n");
					configLogger.warn("Option --malt0.4-behavior = true and --singlemalt-null_value != 'rootlabel'. Option --singlemalt-null_value is overloaded with value 'rootlabel'\n");
				}
				if (getOptionValue("malt0.4", "depset").toString().equals("")) {				
					configLogger.warn("Option --malt0.4-behavior = true and option --malt0.4-depset has no value. These combination will probably not reproduce the behavior of MaltParser 0.4 (C-impl)\n");
				}
				if (getOptionValue("malt0.4", "posset").toString().equals("")) {				
					configLogger.warn("Option --malt0.4-behavior = true and option --malt0.4-posset has no value. These combination will probably not reproduce the behavior of MaltParser 0.4 (C-impl)\n");
				}
				if (getOptionValue("malt0.4", "cposset").toString().equals("")) {				
					configLogger.warn("Option --malt0.4-behavior = true and option --malt0.4-cposset has no value. These combination will probably not reproduce the behavior of MaltParser 0.4 (C-impl)\n");
				}
				if (!getOptionValue("guide", "kbest").toString().equals("1")) {
					MaltConsoleEngine.getOptionManager().overloadOptionValue(containerIndex, "guide", "kbest", "1");
					configDir.getInfoFileWriter().write("--guide-kbest (  -k)                    1\n");
					configLogger.warn("Option --malt0.4-behavior = true and --guide-kbest != '1'. Option --guide-kbest is overloaded with value '1'\n");
				}
			}
			if (getOptionValue("guide", "features").toString().equals("")) {
				MaltConsoleEngine.getOptionManager().overloadOptionValue(containerIndex, "guide", "features", getOptionValueString("singlemalt", "parsing_algorithm"));
				configDir.getInfoFileWriter().write("--guide-features (  -F)                 "+getOptionValue("guide", "features").toString()+"\n");
			}
			if (getOptionValue("guide", "data_split_column").toString().equals("") && !getOptionValue("guide", "data_split_structure").toString().equals("")) {
				configLogger.warn("Option --guide-data_split_column = '' and --guide-data_split_structure != ''. Option --guide-data_split_structure is overloaded with '', this will cause the parser to induce a single model.\n ");
				MaltConsoleEngine.getOptionManager().overloadOptionValue(containerIndex, "guide", "data_split_structure", "");
				configDir.getInfoFileWriter().write("--guide-data_split_structure (  -s)\n");
			}
			if (!getOptionValue("guide", "data_split_column").toString().equals("") && getOptionValue("guide", "data_split_structure").toString().equals("")) {
				configLogger.warn("Option --guide-data_split_column != '' and --guide-data_split_structure = ''. Option --guide-data_split_column is overloaded with '', this will cause the parser to induce a single model.\n");
				MaltConsoleEngine.getOptionManager().overloadOptionValue(containerIndex, "guide", "data_split_column", "");
				configDir.getInfoFileWriter().write("--guide-data_split_column (  -d)\n");
			}
			configDir.getInfoFileWriter().flush();
		} catch (IOException e) {
			throw new ConfigurationException("Could not write to the configuration information file. ", e);
		}
	}
	
	/* (non-Javadoc)
	 * @see se.malt.core.config.Configuration#toString()
	 */
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append(super.toString()+"\n");

		return sb.toString();
	}
}
