package org.maltparser.core.graph;


import org.maltparser.core.dataformat.ColumnDescription;
import org.maltparser.core.exception.MaltChainedException;
import org.maltparser.core.graph.arc.ArcLabels;
import org.maltparser.core.sentence.Sentence;
import org.maltparser.core.symbol.SymbolTable;
/**
 *  
 *
 * @author Johan Hall
 * @since 1.0
**/
public class SingleHeadedDependencyGraph extends DependencyGraph {
	protected boolean acyclic;
	
	
	public SingleHeadedDependencyGraph(Sentence sentence, String rootLabel) throws MaltChainedException {
		super(sentence, rootLabel);
	}
	
	public void addArc(Node head, Node dependent, ArcLabels arcLabelCodes) throws MaltChainedException {
		if (dependent.getIndex().intValue() < 0 && dependent.getIndex().intValue() > getSize()) {
			throw new GraphException("The depedent node is out of bound when adding an arc. ");
		} else if (head.getIndex().intValue() < 0 && head.getIndex().intValue() > getSize()) {
			throw new GraphException("The head node is out of bound when adding an arc. ");
		} 
		Node hc = nodes[head.getIndex()].findComponent();
		Node dc = nodes[dependent.getIndex()].findComponent();
		if (hc != dc) {
			link(hc, dc);
			numberOfComponents--;		
		} else {
			acyclic = false;
		}
		
		nodes[dependent.getIndex()].addHead(nodes[head.getIndex()], arcLabelCodes);
		if (singleheaded == true) {
			singleheaded = nodes[dependent.getIndex()].isSingleHeaded();
		}
	}
	
	protected void initNodes(String rootLabel) throws MaltChainedException {
		if (sentence != null) {
			nodes = new Node[getCapacity()];
			nodes[0] = new Node(0,true,true); 
			rootLabels = new RootLabels(rootLabel, sentence.getDataFormatInstance().getOutputColumnDescriptions());
			for (int i = 1; i < getCapacity(); i++) {
				nodes[i] = new Node(i, false, true, getSentence().getToken(i));
			}

			for (int i = 1; i < getCapacity() - 1; i++) {
				nodes[i].setSuccessor(nodes[i + 1]);
			}
			for (int i = 2; i < getCapacity(); i++) {
				nodes[i].setPredecessor(nodes[i - 1]);
			}
		}
	}
	
	public String toTab() throws MaltChainedException {
		StringBuilder sb = new StringBuilder();
		for (int i = 1; i <= size; i++) {		
			int j=0;
			for (ColumnDescription column : sentence.getDataFormatInstance()) {
				if (j != 0) {
					sb.append('\t');
				}
				if (column.getType() == ColumnDescription.IGNORE) {
					sb.append(column.getDefaultOutput());
				} else if (column.getCategory() == ColumnDescription.INPUT || column.getCategory() == ColumnDescription.ECHO) {
					sb.append(nodes[i].getToken().getColumnInstance(column).getStringValue());
				} else if (column.getCategory() == ColumnDescription.OUTPUT) {
					SymbolTable table = outputColumnMap.get(column.getSymbolTable());
					Integer labelCode = nodes[i].getLabelCode(table);
					if (labelCode == null) {
						sb.append(rootLabels.getRootLabel(table));
					} else if (labelCode.equals(rootLabels.getRootLabelCode(table))) {
						sb.append(rootLabels.getRootLabel(table));
					} else {
						sb.append(table.getSymbolCodeToString(labelCode));
					}
				} else if (column.getCategory() == ColumnDescription.HEAD) {
					sb.append(nodes[i].getHead().getIndex());
				}
				j++;
			}
			sb.append('\n');
		}
		sb.append('\n');
		return sb.toString();
	}
	
	public void clear() {
		super.clear();
		acyclic = true;
	}
	
	public boolean isProjective() {
		return true;
	}
	
	public boolean isForest() {
		return singleheaded && acyclic;
	}
	
	public boolean isAcyclic() {
		return acyclic;
	}
}
