package org.maltparser.parser.guide.feature;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.HashMap;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.maltparser.core.exception.MaltChainedException;
import org.maltparser.core.helper.Util;
import org.maltparser.parser.algorithm.ParsingAlgorithm;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;


/**


@author Johan Hall
@since 1.0
 */
public class FeatureVector extends ArrayList<Feature> implements Serializable {
	public final static long serialVersionUID = 3256444702936019250L;
	private ParsingAlgorithm parsingAlgorithm;

	public FeatureVector(ParsingAlgorithm parsingAlgorithm) { 
		this.parsingAlgorithm = parsingAlgorithm;
	}
	
	 public void addFeatureLine(String line) throws MaltChainedException {
		 Feature feature = null;
		 line = line.trim();
		 int index = line.indexOf('(');
		 if (index == -1) {
			 throw new FeatureException("There is no start parenthesis at line '"+line+"' in the feature specification file. ");
		 }
		 if (!line.endsWith(")")) {
			 throw new FeatureException("There is no end parenthesis at line '"+line+"' in the feature specification file. ");
		 }
		 Class<?> clazz;
		 try {
			clazz = Class.forName("org.maltparser.parser.guide.feature."+line.substring(0, index)+"Feature");
		 } catch (ClassNotFoundException e) {
			 throw new FeatureException("The feature class '"+line.substring(0, index)+"' cannot be found. ", e);
		 }
		 try {
			Class<?>[] argTypes = { org.maltparser.parser.algorithm.ParsingAlgorithm.class, java.lang.String.class };
			Object[] arguments = new Object[2];
			arguments[0] = parsingAlgorithm;
			arguments[1] = line.substring(index+1,line.length()-1);
			Constructor<?> constructor = clazz.getConstructor(argTypes);
			feature = (Feature)constructor.newInstance(arguments);
		} catch (NoSuchMethodException e) {
			throw new FeatureException("The feature '"+clazz.getName()+"' cannot be initialized. ", e);
		} catch (InstantiationException e) {
			throw new FeatureException("The feature '"+clazz.getName()+"' cannot be initialized. ", e);
		} catch (IllegalAccessException e) {
			throw new FeatureException("The feature '"+clazz.getName()+"' cannot be initialized. ", e);
		} catch (InvocationTargetException e) {
			throw new FeatureException("The feature '"+clazz.getName()+"' cannot be initialized. ", e);
		}
		if (feature != null) {
			add(feature);
		} else {
			throw new FeatureException("The feature cannot be found at feature line '"+line+"'. ");
		}
		
	 }
	

	public void readFeatureSpecification(String featureSpecFileName) throws MaltChainedException {
		if (featureSpecFileName.endsWith(".xml")) {
			readFeatureSpecificationXML(featureSpecFileName);
		} else if (featureSpecFileName.endsWith(".par")) {
			readFeatureSpecificationPAR(featureSpecFileName);
		} else {
			throw new FeatureException("The feature specification file '"+featureSpecFileName+"' must have either the file suffix '.xml' or '.par'. ");
		}
		if (((Boolean)parsingAlgorithm.getConfiguration().getOptionValue("malt0.4", "behavior")).booleanValue() == true) {
			malt04Tricks();
		}
	}
	
	private void readFeatureSpecificationXML(String featureSpecFileName) throws MaltChainedException {
		//File specFile = new File(featureSpecFileName);
		
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            Element root = null;
            
            root = db.parse(Util.findURL(featureSpecFileName).openStream()).getDocumentElement();
            /* if (specFile.exists()) {
            	root = db.parse(specFile).getDocumentElement();
            } else {
            	root = db.parse(getClass().getResource(featureSpecFileName).openStream()).getDocumentElement();
            }*/
            if (root == null) {
            	throw new FeatureException("The feature specification file '"+featureSpecFileName+"' cannot be found. ");
            }

    		NodeList featuremodels = root.getElementsByTagName("featuremodel");
            if (featuremodels.getLength() < 1) {
            	throw new FeatureException("The feature specification file '"+featureSpecFileName+"' should at least contain one feature model. ");
            }
            Element featuremodel = (Element)featuremodels.item(0);
    		NodeList features = featuremodel.getElementsByTagName("feature");
            Element feature;
            String featureText;
            for (int j = 0; j < features.getLength(); j++) {
            	feature = (Element)features.item(j); 
            	featureText = feature.getTextContent().trim();
            	if (featureText.length() > 1) {
            		addFeatureLine(featureText);
            	}
            }
        } catch (IOException e) {
        	throw new FeatureException("The feature specification file '"+featureSpecFileName+"' cannot be found. ", e);
        } catch (ParserConfigurationException e) {
        	throw new FeatureException("Problem parsing the file "+featureSpecFileName+". ", e);
        } catch (SAXException e) {
        	throw new FeatureException("Problem parsing the file "+featureSpecFileName+". ", e);
        }
	}
	
	private void readFeatureSpecificationPAR(String featureSpecFileName) throws MaltChainedException {
		HashMap<String,String> old2newName = new HashMap<String,String>();
		old2newName.put("POS", "POSTAG");
		old2newName.put("CPOS", "CPOSTAG");
		old2newName.put("DEP", "DEPREL");
		old2newName.put("LEX", "FORM");
		old2newName.put("LEMMA", "LEMMA");
		old2newName.put("FEATS", "FEATS");

		try {
			File specFile = new File(featureSpecFileName);
			BufferedReader br = null;
			if (specFile.exists()) {
				br = new BufferedReader(new FileReader(specFile));
			} else {
				br = new BufferedReader(new InputStreamReader(getClass().getResource(featureSpecFileName).openStream()));
			}
			if (br != null) {
				String fileLine;
				String items[];
				StringBuilder featureText = new StringBuilder();
				String splitfeats = "";
				while ((fileLine = br.readLine()) != null) {
					featureText.setLength(0);
					splitfeats = "";
					if (fileLine.length() > 1 && !fileLine.trim().substring(0, 2).trim().equals("--")) {
						items = fileLine.split("\t");
						if (items.length < 2) {
							throw new FeatureException("The feature specification file '"+featureSpecFileName+"' must contain at least two columns.");
						}
						if (!(old2newName.containsKey(items[0].trim()) || old2newName.containsValue(items[0].trim()))) {
							throw new FeatureException("Column one in the feature specification file '"+featureSpecFileName+"' contains an unknown value '"+items[0].trim()+"'. ");
						}
						if (items[0].trim().equalsIgnoreCase("DEP") || items[0].trim().equalsIgnoreCase("DEPREL")) {
							featureText.append("OutputColumn(DEPREL, ");
						} else {
							if (old2newName.containsKey(items[0].trim())) {
								featureText.append("InputColumn("+old2newName.get(items[0].trim())+", ");
							} else if (old2newName.containsValue(items[0].trim())) {
								featureText.append("InputColumn("+items[0].trim()+", ");
							}
							if (items[0].trim().equalsIgnoreCase("FEATS")) {
								splitfeats = "Split(";
							}
						}
						if (!(items[1].trim().equalsIgnoreCase("STACK") || items[1].trim().equalsIgnoreCase("INPUT") || items[1].trim().equalsIgnoreCase("CONTEXT"))) {
							throw new FeatureException("Column two in the feature specification file '"+featureSpecFileName+"' should be either 'STACK', 'INPUT' or 'CONTEXT', not '"+items[1].trim()+"'. ");
						}
						int offset = 0;
						if (items.length >= 3) {
							offset = new Integer(Integer.parseInt(items[2]));
						}
						String functionArg = "";
						if (parsingAlgorithm.getName().equals("nivreeager") || parsingAlgorithm.getName().equals("nivrestandard")) {
							if (items[1].trim().equalsIgnoreCase("STACK")) {
								functionArg = "Stack["+offset+"]";
							} else if (items[1].trim().equalsIgnoreCase("INPUT")) {
								functionArg = "Input["+offset+"]";
							} else {
								throw new FeatureException("Column two in the feature specification file '"+featureSpecFileName+"' should not contain the value '"+items[1].trim()+"' in combination with Nivre's algorithm. ");
							}
						} else if (parsingAlgorithm.getName().equals("covnonproj") || parsingAlgorithm.getName().equals("covproj")) {
							if (items[1].trim().equalsIgnoreCase("STACK")) {
								functionArg = "Left["+offset+"]";
							} else if (items[1].trim().equalsIgnoreCase("INPUT")) {
								functionArg = "Right["+offset+"]";
							} else if (items[1].trim().equalsIgnoreCase("CONTEXT")) {
								if (offset >= 0) {
									functionArg = "LeftContext["+offset+"]";
								} else {
									functionArg = "RightContext["+Math.abs(offset + 1)+"]";
								}
							} else {
								throw new FeatureException("Column two in the feature specification file '"+featureSpecFileName+"' should not contain the value '"+items[1].trim()+"' in combination with Covington's algorithm. ");
							}
						}
						int linearOffset = 0;
						int headOffset = 0;
						int depOffset = 0;
						int sibOffset = 0;
						int suffixLength = 0;
						if (items.length >= 4) { linearOffset = new Integer(Integer.parseInt(items[3])); }
						if (items.length >= 5) { headOffset = new Integer(Integer.parseInt(items[4])); }
						if (items.length >= 6) { depOffset = new Integer(Integer.parseInt(items[5])); }
						if (items.length >= 7) { sibOffset = new Integer(Integer.parseInt(items[6])); }
						if (items.length >= 8) { suffixLength = new Integer(Integer.parseInt(items[7])); }
						if (linearOffset < 0) {
							linearOffset = Math.abs(linearOffset);
							for (int i = 0; i < linearOffset; i++) {
								functionArg = "pred("+functionArg+")"; 
							}
						} else if (linearOffset > 0) {
							for (int i = 0; i < linearOffset; i++) {
								functionArg = "succ("+functionArg+")"; 
							}
						} 
						if (headOffset >= 0) {
							for (int i = 0; i < headOffset; i++) {
								functionArg = "head("+functionArg+")"; 
							}
						} else {
							throw new FeatureException("The feature specification file '"+featureSpecFileName+"' should not contain a negative head function value. ");
						}
						if (depOffset < 0) {
							depOffset = Math.abs(depOffset);
							for (int i = 0; i < depOffset; i++) {
								functionArg = "ldep("+functionArg+")"; 
							}
						} else if (depOffset > 0) {
							for (int i = 0; i < depOffset; i++) {
								functionArg = "rdep("+functionArg+")"; 
							}							
						}
						if (sibOffset < 0) {
							sibOffset = Math.abs(sibOffset);
							for (int i = 0; i < sibOffset; i++) {
								functionArg = "lsib("+functionArg+")"; 
							}
						} else if (sibOffset > 0) {
							for (int i = 0; i < sibOffset; i++) {
								functionArg = "rsib("+functionArg+")"; 
							}							
						}
						if (suffixLength != 0) {
							addFeatureLine("Suffix("+featureText.toString()+functionArg+"),"+suffixLength+")");
						}
						if (splitfeats.equals("Split(")) {
							addFeatureLine(splitfeats+featureText.toString()+functionArg+"),\\|)");
						} else {
			            	addFeatureLine(featureText.toString()+functionArg+")");
						}
					}
				}
				br.close();
			}
		} catch (FileNotFoundException e) {
			throw new FeatureException("The feature specification file '"+featureSpecFileName+"' cannot be found. ", e);
		} catch (IOException e) {
			throw new FeatureException("Could not read the feature specification file '"+featureSpecFileName+"'. ", e);
		} catch (NumberFormatException e) {
			throw new FeatureException("The feature specification file '"+featureSpecFileName+"' contains a illegal integer value. ", e);
		}
	}
	
	public void malt04Tricks() {
		ArrayList<Feature> newSortVector = new ArrayList<Feature>();
		
		// MaltParser 0.4 order: POS, DEP, LEX, LEMMA, CPOS, FEATS -> MaltParser 1.0 order: POSTAG, DEPREL, FORM, LEMMA, CPOS, FEATS (SplitFEATS)
		String[] order = new String[]{"InputColumn(POSTAG", "OutputColumn(DEPREL", "InputColumn(FORM", "InputColumn(LEMMA", "InputColumn(CPOSTAG", "InputColumn(FEATS", "Split(InputColumn(FEATS"};
		for (int i=0; i<order.length;i++) {
			for (int j=0; j<size();j++) {
				if (get(j).toString().startsWith(order[i])) {
					newSortVector.add(get(j));
				}
			}
		}
		// Replace the feature vector with a new feature vector according to the order of MaltParser 0.4
		clear();
		for (int i=0; i<newSortVector.size();i++) {
			add(newSortVector.get(i));
		}
		//Replace rdep with rdep2 (that cannot return the right dependent of the root node)
		for (int j=0; j<size();j++) {
			TokenFeature cf = null;
			if (get(j) instanceof TokenFeature) {
				cf = (TokenFeature)get(j);
			} else if (get(j) instanceof SplitFeature && ((SplitFeature)get(j)).getParentFeature() instanceof TokenFeature) {
				cf = (TokenFeature)((SplitFeature)get(j)).getParentFeature();
			}  else if (get(j) instanceof SuffixFeature && ((SuffixFeature)get(j)).getParentFeature() instanceof TokenFeature) {
				cf = (TokenFeature)((SuffixFeature)get(j)).getParentFeature();
			}
			if (cf != null) {
				cf.replaceStackItem("rdep", "rdep2");
			}
		}
	}
	
	
	public String toString() {
		StringBuilder sb = new StringBuilder();
		for (Feature feature : this) {
			sb.append(feature+"\n");
		}
		return sb.toString();
	}
}
