package org.maltparser.transform;

import java.util.Vector;

import org.apache.log4j.Logger;
import org.maltparser.core.exception.MaltChainedException;
import org.maltparser.core.graph.DependencyGraph;
import org.maltparser.core.graph.Node;
import org.maltparser.core.symbol.SymbolTable;

/**
 * This class contains methods for projectivizing and deprojectivizing
 * 
 * @author Jens Nilsson
 * @since 1.0
 */
public class PseudoProjectivity {
	static int id = 0;

	private enum PseudoProjectiveEncoding {
		NONE, BASELINE, HEAD, PATH, HEADPATH, TRACE
	};

	private enum CoveredRootAttachment {
		NONE, LEFT, RIGHT, HEAD
	};

	private enum LiftingOrder {
		SHORTEST, DEEPEST
	};

	private PseudoProjectiveEncoding markingStrategy;
	private CoveredRootAttachment rootAttachment;
	private LiftingOrder liftingOrder;
	private Logger configLogger;

	private SymbolTable deprelSymbolTable;

	private Vector<Boolean> nodeProjectivity;
	private Vector<Integer> nodeDepth;
	private Vector<Boolean> nodeLifted;
	private Vector<Vector<Node>> nodeTrace;
	private Vector<Node> headDeprel;
	private Vector<Boolean> nodePath;
	private Vector<Boolean> isCoveredRoot;
	private Vector<String> originalDeprel;

	private Vector<Integer> nodeRelationLength;

	private Vector<String> synacticHeadDeprel;

	private boolean areNodesProjective;

	public PseudoProjectivity(String markingStrategy, String coveredRoot, String liftingOrder, Logger configLogger,
			SymbolTable deprelSymbolTable) throws MaltChainedException {
		nodeProjectivity = new Vector<Boolean>();
		nodeDepth = new Vector<Integer>();
		nodeLifted = new Vector<Boolean>();
		nodeTrace = new Vector<Vector<Node>>();
		headDeprel = new Vector<Node>();
		nodePath = new Vector<Boolean>();
		isCoveredRoot = new Vector<Boolean>();
		originalDeprel = new Vector<String>();

		nodeRelationLength = new Vector<Integer>();

		synacticHeadDeprel = new Vector<String>();

		this.deprelSymbolTable = deprelSymbolTable;
		this.configLogger = configLogger;
		if (markingStrategy.equalsIgnoreCase("none")) {
			this.markingStrategy = PseudoProjectiveEncoding.NONE;
		} else if (markingStrategy.equalsIgnoreCase("baseline")) {
			this.markingStrategy = PseudoProjectiveEncoding.BASELINE;
		} else if (markingStrategy.equalsIgnoreCase("head")) {
			this.markingStrategy = PseudoProjectiveEncoding.HEAD;
		} else if (markingStrategy.equalsIgnoreCase("path")) {
			this.markingStrategy = PseudoProjectiveEncoding.PATH;
		} else if (markingStrategy.equalsIgnoreCase("head+path")) {
			this.markingStrategy = PseudoProjectiveEncoding.HEADPATH;
		} else if (markingStrategy.equalsIgnoreCase("trace")) {
			this.markingStrategy = PseudoProjectiveEncoding.TRACE;
		}

		if (coveredRoot.equalsIgnoreCase("none")) {
			this.rootAttachment = CoveredRootAttachment.NONE;
		} else if (coveredRoot.equalsIgnoreCase("left")) {
			this.rootAttachment = CoveredRootAttachment.LEFT;
		} else if (coveredRoot.equalsIgnoreCase("right")) {
			this.rootAttachment = CoveredRootAttachment.RIGHT;
		} else if (coveredRoot.equalsIgnoreCase("head")) {
			this.rootAttachment = CoveredRootAttachment.HEAD;
		}

		if (liftingOrder.equalsIgnoreCase("shortest")) {
			this.liftingOrder = LiftingOrder.SHORTEST;
		} else if (liftingOrder.equalsIgnoreCase("deepest")) {
			this.liftingOrder = LiftingOrder.DEEPEST;
		}
	}

	private void initProjectivization(DependencyGraph pdg) throws MaltChainedException {
		int i = 0;
		nodeLifted.clear();
		nodeTrace.clear();
		headDeprel.clear();
		nodePath.clear();
		isCoveredRoot.clear();
		originalDeprel.clear();

		nodeRelationLength.clear();

		for (i = 0; i <= pdg.getSize(); i++) {
			nodeLifted.add(false);
			nodeTrace.add(new Vector<Node>());
			headDeprel.add(null);
			nodePath.add(false);
			isCoveredRoot.add(false);
			if (i == 0) {
				originalDeprel.add(null);
			} else {
				originalDeprel.add(deprelSymbolTable.getSymbolCodeToString(pdg.getNode(i).getLabelCode(deprelSymbolTable)));
			}
		}
		areNodesProjective = true;
		computeRelationLength(pdg);
	}

	public void projectivize(DependencyGraph pdg) throws MaltChainedException {
		int i;
		Node deepestNonProjectiveNode;
		id++;
		initProjectivization(pdg);
		if (markingStrategy != PseudoProjectiveEncoding.NONE) {
			while (!isProjective(pdg)) {
				if (liftingOrder == LiftingOrder.DEEPEST) {
					deepestNonProjectiveNode = getDeepestNonProjectiveNode(pdg);
				} else {
					deepestNonProjectiveNode = getShortestNonProjectiveNode(pdg);
				}
				if (!attachCoveredRoots(pdg, deepestNonProjectiveNode)) {
					nodeLifted.set(deepestNonProjectiveNode.getIndex(), true);
					setHeadDeprel(deepestNonProjectiveNode, deepestNonProjectiveNode.getHead());
					setPath(deepestNonProjectiveNode.getHead());
					pdg.addArc(pdg.getNode(deepestNonProjectiveNode.getHead().getHead().getIndex()), deepestNonProjectiveNode,
							deepestNonProjectiveNode.getLabelCode());
				}
			}
			if (rootAttachment == CoveredRootAttachment.NONE) {
				deattachCoveredRootsForProjectivization(pdg);
			}
		} else if (rootAttachment != CoveredRootAttachment.NONE) {
			for (i = 1; i < pdg.getSentence().getSize(); i++) {
				attachCoveredRoots(pdg, pdg.getNode(i));
			}
		}
		// collectTraceStatistics(pdg);
		assignPseudoProjectiveDeprels(pdg);
	}

	private void setHeadDeprel(Node node, Node parent) {
		if (headDeprel.get(node.getIndex()) == null) {
			headDeprel.set(node.getIndex(), parent);
		}
		nodeTrace.set(node.getIndex(), headDeprel);
	}

	private void setPath(Node node) {
		nodePath.set(node.getIndex(), true);
	}

	private boolean isCoveredRoot(Node node) {
		return isCoveredRoot.get(node.getIndex());
	}

	private void deattachCoveredRootsForProjectivization(DependencyGraph pdg) throws MaltChainedException {
		int i, labelIndex;
		String newLabel;
		for (i = 1; i <= pdg.getSize(); i++) {
			if (isCoveredRoot(pdg.getNode(i))) {
				labelIndex = deprelSymbolTable.getSymbolCodeToString(pdg.getNode(i).getLabelCode(deprelSymbolTable)).indexOf("|null");
				newLabel = deprelSymbolTable.getSymbolCodeToString(pdg.getNode(i).getLabelCode(deprelSymbolTable)).substring(0, labelIndex);
				addArc(pdg, pdg.getNode(i), pdg.getRoot(), newLabel);
			}
		}
	}

	private int addSymbol(String newLabel) throws MaltChainedException {
		return deprelSymbolTable.addSymbol(newLabel);
	}

	private void addArc(DependencyGraph pdg, Node node, Node parent, String newLabel) throws MaltChainedException {
		node.getLabelCode().clear();
		node.getLabelCode().put(deprelSymbolTable, addSymbol(newLabel));
		pdg.addArc(parent, node, node.getLabelCode());
	}

	private boolean attachCoveredRoots(DependencyGraph pdg, Node deepest) throws MaltChainedException {
		int i;
		boolean foundCoveredRoot = false;
		Node coveredRootHead;

		for (i = Math.min(deepest.getIndex(), deepest.getHead().getIndex()) + 1; i < Math.max(deepest.getIndex(), deepest.getHead()
				.getIndex()); i++) {
			if (pdg.getNode(i).getHead().isRoot() && !deepest.getHead().isRoot()
					&& Math.min(deepest.getIndex(), deepest.getHead().getIndex()) < pdg.getNode(i).getLeftmostDescendant().getIndex()
					&& pdg.getNode(i).getRightmostDescendant().getIndex() < Math.max(deepest.getIndex(), deepest.getHead().getIndex())) {
				if (rootAttachment == CoveredRootAttachment.LEFT) {
					if (deepest.getHead().getIndex() < deepest.getIndex()) {
						coveredRootHead = deepest.getHead();
					} else {
						coveredRootHead = deepest;
					}
				} else if (rootAttachment == CoveredRootAttachment.RIGHT) {
					if (deepest.getIndex() < deepest.getHead().getIndex()) {
						coveredRootHead = deepest.getHead();
					} else {
						coveredRootHead = deepest;
					}
				} else {
					coveredRootHead = deepest.getHead();
				}
				addArc(pdg, pdg.getNode(i), coveredRootHead, deprelSymbolTable.getSymbolCodeToString(pdg.getNode(i).getLabelCode(
						deprelSymbolTable))
						+ "|null");
				setCoveredRoot(pdg.getNode(i));
				foundCoveredRoot = true;
			}
		}
		return foundCoveredRoot;
	}

	private void setCoveredRoot(Node node) {
		isCoveredRoot.set(node.getIndex(), true);
	}

	public boolean isProjective(DependencyGraph pdg) throws MaltChainedException {
		int i = 0;
		nodeProjectivity.clear();
		nodeDepth.clear();
		for (i = 0; i <= pdg.getSize(); i++) {
			nodeProjectivity.add(true);
			nodeDepth.add(-1);
		}
		areNodesProjective = true;
		isProjective(pdg, pdg.getRoot());
		return areNodesProjective;
	}

	private Vector<Node> isProjective(DependencyGraph pdg, Node node) {
		int i, j;
		Vector<Vector<Node>> leftChildren = new Vector<Vector<Node>>();
		Vector<Vector<Node>> rightChildren = new Vector<Vector<Node>>();
		Vector<Node> toReturn = new Vector<Node>();
		Node child;

		toReturn.add(node);
		i = 0;
		while ((child = node.getLeftDependent(i)) != null) {
			leftChildren.add(isProjective(pdg, child));
			toReturn.addAll(leftChildren.lastElement());
			i++;
		}
		i = 0;
		while ((child = node.getRightDependent(i)) != null) {
			rightChildren.add(isProjective(pdg, child));
			toReturn.addAll(rightChildren.lastElement());
			i++;
		}
		boolean[] visited = new boolean[pdg.getSize()];
		for (i = 0; i < leftChildren.size(); i++) {
			Vector<Node> dominating = leftChildren.get(i);
			for (Node descendant : dominating) {
				visited[descendant.getIndex() - 1] = true;
			}
		}
		for (i = 0; i < rightChildren.size(); i++) {
			Vector<Node> dominating = rightChildren.get(i);
			for (Node descendant : dominating) {
				visited[descendant.getIndex() - 1] = true;
			}
		}

		for (i = 0; i < leftChildren.size(); i++) {
			for (j = node.getLeftDependent(i).getIndex(); j < node.getIndex(); j++) {
				if (!visited[j - 1]) {
					areNodesProjective = false;
					nodeProjectivity.set(node.getLeftDependent(i).getIndex(), false);
				}
			}
		}
		for (i = 0; i < rightChildren.size(); i++) {
			for (j = node.getIndex() + 1; j < node.getRightDependent(i).getIndex(); j++) {
				if (j != 0 && !visited[j - 1]) {
					areNodesProjective = false;
					nodeProjectivity.set(node.getRightDependent(i).getIndex(), false);
				}
			}
		}
		return toReturn;
	}

	private Node getDeepestNonProjectiveNode(DependencyGraph pdg) throws MaltChainedException {
		int i;
		Node deepestNonProjectiveNode = null;
		computeNodeDepth(pdg.getRoot(), 0);
		for (i = 0; i <= pdg.getSize(); i++) {
			if (!nodeProjectivity.get(i)
					&& (deepestNonProjectiveNode == null || nodeDepth.get(i) > nodeDepth.get(deepestNonProjectiveNode.getIndex()))) {
				deepestNonProjectiveNode = pdg.getNode(i);
			}
		}
		return deepestNonProjectiveNode;
	}

	private Node getShortestNonProjectiveNode(DependencyGraph pdg) throws MaltChainedException {
		int i;
		Node shortestNonProjectiveNode = null;
		for (i = 0; i <= pdg.getSize(); i++) {
			if (!nodeProjectivity.get(i)
					&& (shortestNonProjectiveNode == null
							|| nodeRelationLength.get(i) < nodeRelationLength.get(shortestNonProjectiveNode.getIndex()) || (nodeRelationLength
							.get(i) == nodeRelationLength.get(shortestNonProjectiveNode.getIndex())))) {
				shortestNonProjectiveNode = pdg.getNode(i);
			}
		}
		return shortestNonProjectiveNode;
	}

	private void computeNodeDepth(Node node, int depth) {
		int i;
		Node child;

		nodeDepth.set(node.getIndex(), depth);
		i = 0;
		while ((child = node.getLeftDependent(i)) != null) {
			computeNodeDepth(child, depth + 1);
			i++;
		}
		i = 0;
		while ((child = node.getRightDependent(i)) != null) {
			computeNodeDepth(child, depth + 1);
			i++;
		}
	}

	private void computeRelationLength(DependencyGraph pdg) {
		int i;
		nodeRelationLength.add(0);
		for (i = 1; i <= pdg.getSize(); i++) {
			nodeRelationLength.add(Math.abs(pdg.getNode(i).getIndex() - pdg.getNode(i).getHead().getIndex()));
		}
	}

	private void assignPseudoProjectiveDeprels(DependencyGraph pdg) throws MaltChainedException {
		int i;
		String newLabel;
		for (i = 1; i <= pdg.getSize(); i++) {
			if (!isCoveredRoot.get(i)) {
				if (markingStrategy == PseudoProjectiveEncoding.HEAD) {
					if (nodeLifted.get(i)) {
						newLabel = deprelSymbolTable.getSymbolCodeToString(pdg.getNode(i).getLabelCode(deprelSymbolTable)) + "|"
								+ originalDeprel.get(headDeprel.get(i).getIndex());
					} else {
						newLabel = deprelSymbolTable.getSymbolCodeToString(pdg.getNode(i).getLabelCode(deprelSymbolTable));
					}
					addArc(pdg, pdg.getNode(i), pdg.getNode(i).getHead(), newLabel);
				} else if (markingStrategy == PseudoProjectiveEncoding.PATH) {
					if (nodeLifted.get(i) && nodePath.get(i)) {
						newLabel = deprelSymbolTable.getSymbolCodeToString(pdg.getNode(i).getLabelCode(deprelSymbolTable)) + "|%";
						addArc(pdg, pdg.getNode(i), pdg.getNode(i).getHead(), newLabel);
					} else if (nodeLifted.get(i) && !nodePath.get(i)) {
						newLabel = deprelSymbolTable.getSymbolCodeToString(pdg.getNode(i).getLabelCode(deprelSymbolTable)) + "|";
						addArc(pdg, pdg.getNode(i), pdg.getNode(i).getHead(), newLabel);
					} else if (!nodeLifted.get(i) && nodePath.get(i)) {
						newLabel = deprelSymbolTable.getSymbolCodeToString(pdg.getNode(i).getLabelCode(deprelSymbolTable)) + "%";
						addArc(pdg, pdg.getNode(i), pdg.getNode(i).getHead(), newLabel);
					}
				} else if (markingStrategy == PseudoProjectiveEncoding.HEADPATH) {
					if (nodeLifted.get(i) && nodePath.get(i)) {
						newLabel = deprelSymbolTable.getSymbolCodeToString(pdg.getNode(i).getLabelCode(deprelSymbolTable)) + "|"
								+ originalDeprel.get(headDeprel.get(i).getIndex()) + "%";
						addArc(pdg, pdg.getNode(i), pdg.getNode(i).getHead(), newLabel);
					} else if (nodeLifted.get(i) && !nodePath.get(i)) {
						newLabel = deprelSymbolTable.getSymbolCodeToString(pdg.getNode(i).getLabelCode(deprelSymbolTable)) + "|"
								+ originalDeprel.get(headDeprel.get(i).getIndex());
						addArc(pdg, pdg.getNode(i), pdg.getNode(i).getHead(), newLabel);
					} else if (!nodeLifted.get(i) && nodePath.get(i)) {
						newLabel = originalDeprel.get(pdg.getNode(i).getIndex()) + "%";
						addArc(pdg, pdg.getNode(i), pdg.getNode(i).getHead(), newLabel);
					}
				} else if (markingStrategy == PseudoProjectiveEncoding.TRACE) {
					if (nodeLifted.get(i)) {
						newLabel = deprelSymbolTable.getSymbolCodeToString(pdg.getNode(i).getLabelCode(deprelSymbolTable)) + "|";
						addArc(pdg, pdg.getNode(i), pdg.getNode(i).getHead(), newLabel);
					}
				}
			}
		}
	}

	public void deprojectivize(DependencyGraph pdg) throws MaltChainedException {
		int i;
		int pathLabelIndex = -1, movedLabelIndex = -1;
		String label;
		initDeprojeciviztion(pdg);

		for (i = 1; i <= pdg.getSize(); i++) {
			if (pdg.getNode(i).getLabelCode(deprelSymbolTable) != null) {
				label = deprelSymbolTable.getSymbolCodeToString(pdg.getNode(i).getLabelCode(deprelSymbolTable));
				if (label != null && (pathLabelIndex = label.indexOf("%")) != -1) {
					setPath(pdg.getNode(i));
					label = label.substring(0, pathLabelIndex);
				}
				if (label != null && (movedLabelIndex = label.indexOf("|")) != -1 && label.indexOf("|null") == -1) {
					nodeLifted.set(i, true);
					if (movedLabelIndex + 1 < label.length()) {
						synacticHeadDeprel.set(i, label.substring(movedLabelIndex + 1));
					}
					label = label.substring(0, movedLabelIndex);
				}
				if (pathLabelIndex != -1 || movedLabelIndex != -1) {
					addArc(pdg, pdg.getNode(i), pdg.getNode(pdg.getNode(i).getHead().getIndex()), label);
				}
			}
		}
		deattachCoveredRootsForDerojectivization(pdg);
		if (markingStrategy == PseudoProjectiveEncoding.HEAD) {
			deprojectivizeWithHead(pdg, pdg.getRoot());
		} else if (markingStrategy == PseudoProjectiveEncoding.PATH) {
			deprojectivizeWithPath(pdg, pdg.getRoot());
		} else if (markingStrategy == PseudoProjectiveEncoding.HEADPATH) {
			deprojectivizeWithHeadAndPath(pdg, pdg.getRoot());
			// } else if (markingStrategy == PseudoProjectiveEncoding.TRACE) {
			// deprojectivizeWithTrace(pdg, pdg.getRoot());
		}
	}

	private void initDeprojeciviztion(DependencyGraph pdg) {
		int i;

		nodeLifted.clear();
		nodePath.clear();
		synacticHeadDeprel.clear();
		for (i = 0; i <= pdg.getSize(); i++) {
			nodeLifted.add(false);
			nodePath.add(false);
			synacticHeadDeprel.add(null);
		}
	}

	private void deattachCoveredRootsForDerojectivization(DependencyGraph pdg) throws MaltChainedException {
		int i, labelIndex;
		String oldLabel, newLabel;
		for (i = 1; i <= pdg.getSize(); i++) {
			if (pdg.getNode(i).getLabelCode(deprelSymbolTable) != null) {
				oldLabel = deprelSymbolTable.getSymbolCodeToString(pdg.getNode(i).getLabelCode(deprelSymbolTable));
				if ((labelIndex = oldLabel.indexOf("|null")) != -1) {
					newLabel = oldLabel.substring(0, labelIndex);
					addArc(pdg, pdg.getNode(i), pdg.getRoot(), newLabel);
				}
			}
		}
	}

	private boolean deprojectivizeWithHead(DependencyGraph pdg, Node node) throws MaltChainedException {
		boolean success = true, childSuccess = false;
		int i, childAttempts = 2;
		Node child, possibleSyntacticHead;
		String nonProjDeprel, syntacticHeadDeprel;
		if (nodeLifted.get(node.getIndex())) {
			syntacticHeadDeprel = synacticHeadDeprel.get(node.getIndex());
			nonProjDeprel = deprelSymbolTable.getSymbolCodeToString(node.getLabelCode(deprelSymbolTable));
			possibleSyntacticHead = breadthFirstSearchSortedByDistanceForHead(pdg, node.getHead(), node, syntacticHeadDeprel);
			if (possibleSyntacticHead != null) {
				addArc(pdg, node, possibleSyntacticHead, nonProjDeprel);
				nodeLifted.set(node.getIndex(), false);
			} else {
				success = false;
			}
		}
		while (!childSuccess && childAttempts > 0) {
			childSuccess = true;
			Vector<Node> children = new Vector<Node>();
			i = 0;
			while ((child = node.getLeftDependent(i)) != null) {
				children.add(child);
				i++;
			}
			i = 0;
			while ((child = node.getRightDependent(i)) != null) {
				children.add(child);
				i++;
			}
			for (i = 0; i < children.size(); i++) {
				child = children.get(i);
				if (!deprojectivizeWithHead(pdg, child)) {
					childSuccess = false;
				}
			}
			childAttempts--;
		}
		return childSuccess && success;
	}

	private Node breadthFirstSearchSortedByDistanceForHead(DependencyGraph dg, Node start, Node avoid, String syntacticHeadDeprel)
			throws MaltChainedException {
		Node dependent;
		String dependentDeprel;
		Vector<Node> nodes = new Vector<Node>();
		nodes.addAll(findAllDependentsVectorSortedByDistanceToPProjNode(dg, start, avoid, false));
		while (nodes.size() > 0) {
			dependent = nodes.remove(0);
			dependentDeprel = deprelSymbolTable.getSymbolCodeToString(dependent.getLabelCode(deprelSymbolTable));
			if (dependentDeprel.equals(syntacticHeadDeprel)) {
				return dependent;
			}
			nodes.addAll(findAllDependentsVectorSortedByDistanceToPProjNode(dg, dependent, avoid, false));
		}
		return null;
	}

	private Vector<Node> findAllDependentsVectorSortedByDistanceToPProjNode(DependencyGraph dg, Node governor, Node avoid,
			boolean percentOnly) {
		int i, j;
		Vector<Node> dependents = new Vector<Node>();
		Node leftChild, rightChild;

		i = governor.getLeftDependentCount() - 1;
		j = 0;
		leftChild = governor.getLeftDependent(i--);
		rightChild = governor.getRightDependent(j++);

		while (leftChild != null && rightChild != null) {
			if (leftChild == avoid) {
				leftChild = governor.getLeftDependent(i--);
			} else if (rightChild == avoid) {
				rightChild = governor.getRightDependent(j++);
			} else if (Math.abs(leftChild.getIndex() - avoid.getIndex()) < Math.abs(rightChild.getIndex() - avoid.getIndex())) {
				if (!percentOnly || (percentOnly && nodePath.get(leftChild.getIndex()))) {
					dependents.add(leftChild);
				}
				leftChild = governor.getLeftDependent(i--);
			} else {
				if (!percentOnly || (percentOnly && nodePath.get(rightChild.getIndex()))) {
					dependents.add(rightChild);
				}
				rightChild = governor.getRightDependent(j++);
			}
		}
		while (leftChild != null) {
			if (leftChild != avoid && (!percentOnly || (percentOnly && nodePath.get(leftChild.getIndex())))) {
				dependents.add(leftChild);
			}
			leftChild = governor.getLeftDependent(i--);
		}
		while (rightChild != null) {
			if (rightChild != avoid && (!percentOnly || (percentOnly && nodePath.get(rightChild.getIndex())))) {
				dependents.add(rightChild);
			}
			rightChild = governor.getRightDependent(j++);
		}
		return dependents;
	}

	private boolean deprojectivizeWithPath(DependencyGraph pdg, Node node) throws MaltChainedException {
		boolean success = true, childSuccess = false;
		int i, childAttempts = 2;
		Node child, possibleSyntacticHead;
		String nonProjDeprel;
		if (node.getLabelCode() != null && nodeLifted.get(node.getIndex()) && nodePath.get(node.getIndex())) {
			nonProjDeprel = deprelSymbolTable.getSymbolCodeToString(node.getLabelCode(deprelSymbolTable));
			possibleSyntacticHead = breadthFirstSearchSortedByDistanceForPath(pdg, node.getHead(), node);
			if (possibleSyntacticHead != null) {
				addArc(pdg, node, possibleSyntacticHead, nonProjDeprel);
				nodeLifted.set(node.getIndex(), false);
			} else {
				success = false;
			}
		}
		if (node.getLabelCode() != null && nodeLifted.get(node.getIndex())) {
			nonProjDeprel = deprelSymbolTable.getSymbolCodeToString(node.getLabelCode(deprelSymbolTable));
			possibleSyntacticHead = breadthFirstSearchSortedByDistanceForPath(pdg, node.getHead(), node);
			if (possibleSyntacticHead != null) {
				addArc(pdg, node, possibleSyntacticHead, nonProjDeprel);
				nodeLifted.set(node.getIndex(), false);
			} else {
				success = false;
			}
		}
		while (!childSuccess && childAttempts > 0) {
			childSuccess = true;
			Vector<Node> children = new Vector<Node>();
			i = 0;
			while ((child = node.getLeftDependent(i)) != null) {
				children.add(child);
				i++;
			}
			i = 0;
			while ((child = node.getRightDependent(i)) != null) {
				children.add(child);
				i++;
			}
			for (i = 0; i < children.size(); i++) {
				child = children.get(i);
				if (!deprojectivizeWithPath(pdg, child)) {
					childSuccess = false;
				}
			}
			childAttempts--;
		}
		return childSuccess && success;
	}

	private Node breadthFirstSearchSortedByDistanceForPath(DependencyGraph dg, Node start, Node avoid) {
		Node dependent;
		Vector<Node> nodes = new Vector<Node>(), newNodes;
		nodes.addAll(findAllDependentsVectorSortedByDistanceToPProjNode(dg, start, avoid, true));
		while (nodes.size() > 0) {
			dependent = nodes.remove(0);
			if (((newNodes = findAllDependentsVectorSortedByDistanceToPProjNode(dg, dependent, avoid, true)).size()) == 0) {
				return dependent;
			}
			nodes.addAll(newNodes);
		}
		return null;
	}

	private boolean deprojectivizeWithHeadAndPath(DependencyGraph dg, Node node) throws MaltChainedException {
		boolean success = true, childSuccess = false;
		int i, childAttempts = 2;
		Node child, possibleSyntacticHead;
		String nonProjDeprelCode;
		if (node.getLabelCode() != null && nodeLifted.get(node.getIndex()) && nodePath.get(node.getIndex())) {
			nonProjDeprelCode = deprelSymbolTable.getSymbolCodeToString(node.getLabelCode(deprelSymbolTable));
			possibleSyntacticHead = breadthFirstSearchSortedByDistanceForHeadAndPath(dg, node.getHead(), node, synacticHeadDeprel.get(node
					.getIndex()));
			if (possibleSyntacticHead != null) {
				addArc(dg, node, possibleSyntacticHead, nonProjDeprelCode);
				nodeLifted.set(node.getIndex(), false);
			} else {
				success = false;
			}
		}
		if (node.getLabelCode() != null && nodeLifted.get(node.getIndex())) {
			nonProjDeprelCode = deprelSymbolTable.getSymbolCodeToString(node.getLabelCode(deprelSymbolTable));
			possibleSyntacticHead = breadthFirstSearchSortedByDistanceForHeadAndPath(dg, node.getHead(), node, synacticHeadDeprel.get(node
					.getIndex()));
			if (possibleSyntacticHead != null) {
				addArc(dg, node, possibleSyntacticHead, nonProjDeprelCode);
				nodeLifted.set(node.getIndex(), false);
			} else {
				success = false;
			}
		}
		while (!childSuccess && childAttempts > 0) {
			childSuccess = true;
			Vector<Node> children = new Vector<Node>();
			i = 0;
			while ((child = node.getLeftDependent(i)) != null) {
				children.add(child);
				i++;
			}
			i = 0;
			while ((child = node.getRightDependent(i)) != null) {
				children.add(child);
				i++;
			}
			for (i = 0; i < children.size(); i++) {
				child = children.get(i);
				if (!deprojectivizeWithHeadAndPath(dg, child)) {
					childSuccess = false;
				}
			}
			childAttempts--;
		}
		return childSuccess && success;
	}

	private Node breadthFirstSearchSortedByDistanceForHeadAndPath(DependencyGraph dg, Node start, Node avoid, String syntacticHeadDeprelCode)
			throws MaltChainedException {
		Node dependent;
		Vector<Node> nodes = new Vector<Node>(), newNodes = null, secondChance = new Vector<Node>();
		nodes.addAll(findAllDependentsVectorSortedByDistanceToPProjNode(dg, start, avoid, true));
		while (nodes.size() > 0) {
			dependent = nodes.remove(0);
			if (((newNodes = findAllDependentsVectorSortedByDistanceToPProjNode(dg, dependent, avoid, true)).size()) == 0
					&& deprelSymbolTable.getSymbolCodeToString(dependent.getLabelCode(deprelSymbolTable)).equals(syntacticHeadDeprelCode)) {
				return dependent;
			}
			nodes.addAll(newNodes);
			if (deprelSymbolTable.getSymbolCodeToString(dependent.getLabelCode(deprelSymbolTable)).equals(syntacticHeadDeprelCode)
					&& newNodes.size() != 0) {
				secondChance.add(dependent);
			}
		}
		if (secondChance.size() > 0) {
			return secondChance.firstElement();
		}
		return null;
	}
}
